import numpy as np 
from ExtraTree.ExtraEnsemble import StandardForestRegressor, ExtraForestRegressor


def test_standard_forest_regressor():
    
    for splitter in ["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"]:
        for threshold in [0, 0.01]:
            for parallel_jobs in [0, 5]:
                for max_features in [0.9, 1]:
                    for max_samples in [0.9, 1]:
                
                        np.random.seed(666)
                        X_train = np.random.rand(200).reshape(-1,2)
                        X_test = np.random.rand(200).reshape(-1,2)
                        Y_train = np.ones(100)
        
                        model = StandardForestRegressor( n_estimators = 20,
                                                max_features = max_features,
                                                max_samples = max_samples,
                                                ensemble_parallel = int(5-parallel_jobs),
                                                splitter = splitter,
                                                min_samples_split = 5, 
                                                min_samples_leaf = 2,
                                                max_depth = 2, 
                                                log_Xrange = True, 
                                                random_state = 666,
                                                parallel_jobs = parallel_jobs, 
                                                search_number = 10,
                                                threshold = threshold)
                        model.fit(X_train, Y_train)
                        assert ((model.predict(X_test)-1)**2).mean()<0.03
    
    
def test_extra_forest_regressor():
    
    for splitter in ["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"]:
        for order in [0,1]:
            for threshold in [0,0.01]:
                for parallel_jobs in [0,5]:
                    for lamda in [0.0001]:
                        for max_features in [0.9, 1]:
                            for max_samples in [0.9, 1]:
                    
                                np.random.seed(666)
                                X_train = np.random.rand(200).reshape(-1,2)
                                X_test = np.random.rand(200).reshape(-1,2)
                                Y_train = np.ones(100)
        
                                model = ExtraForestRegressor( n_estimators = 20,
                                                        max_features = max_features,
                                                        max_samples = max_samples,
                                                        ensemble_parallel = int(5-parallel_jobs),
                                                        splitter = splitter,
                                                        min_samples_split = 5, 
                                                        min_samples_leaf = 2,
                                                        max_depth = 2, 
                                                        order = order, 
                                                        log_Xrange = True, 
                                                        random_state = 666,
                                                        parallel_jobs = parallel_jobs, 
                                                        V = 10,
                                                        r_range_low = 0,
                                                        r_range_up = 1,
                                                        lamda = lamda, 
                                                        search_number = 10,
                                                        threshold = threshold)
                                model.fit(X_train, Y_train)
        
                              
                                assert ((model.predict(X_test)-1)**2).mean()<0.03
                    
                    



import numpy as np
from ExtraTree import StandardTreeRegressor, ExtraTreeRegressor
from sklearn.metrics import mean_squared_error as MSE
from multiprocessing import Pool


BASE_LEARNER = {
    "standard_tree_regressor": StandardTreeRegressor,
    "extra_tree_regressor": ExtraTreeRegressor
    }



def pred_parallel(input_tuple):
    """Parallel instrumental function for prediction. 

    Parameters
    ----------
    input_tuple : (ree object, array of (n_samples_, dim))
        The tuple for parallelization.
    
    Returns
    -------
    prediction : array-like of shape (n_sample_, ) 
        Prediction.
    """
    tree, X = input_tuple
    return tree.predict(X)


class BaseGradientBoosting(object):
    """ Abstact Boosting Structure.
    
    
    Parameters
    ----------
    n_estimators : int
        Number of base learners.
    
    max_samples : float
        Proportion of samples to be bootstrapped, can be larger than 1. 
        
    rho : float
        The learning rate. 
        
    ensemble_parallel : int
        If 0, no parallel. If positive, the parallization is conducted in 
        ensemble_parallel threads.
    
    splitter : splitter keyword in SPLITTERS
        Splitting scheme
        
    estimator : estimator keyword in ESTIMATORS
        Estimation scheme
        
    min_samples_split : int
        The minimum number of samples required to split an internal node.
    
    min_samples_leaf : int
        The minimum number of samples required in the subnodes to split an 
        internal node.
    
    max_depth : int
        Maximum depth of the individual regression estimators.
        
    order : int > 0
        Extrapolation order.
    
    log_Xrange : bool
        If True, the points in each cell is recorded. 
        
    random_state : int
        Random state for building the tree.
        
    parallel_jobs : int
        If 0, no parallel for base learners. If positive, the parallization is 
        conducted in parallel_jobs threads. Can not be positive with ensemble_parallel
        simutanously. 
        
    V : int or "auto"
        Parameter for homothetic estimation. If int, the estimations are taken at 
        i/V, i = 1, ..., V. If auto, it is set to max(n_samples * 2^(- 2 - max_depth), 5). 
        
    r_range_low : float in [0, 1]
        Lower bound of homothetic ratio to consider.
    
    r_range_up : float in [0, 1], > r_range_low
        Upper bound of homothetic ratio to consider.        
    
    lamda : float in [0,infty)
        Ridge regularization parameter. 
        
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    Attributes
    ----------
    trees : list
        List of base learners. 
    
    """
    def __init__(self,  n_estimators = 20, 
                 max_features = 1.0, 
                 max_samples = 1.0,
                 rho = 0.1,
                 ensemble_parallel = 0,
                 splitter = "maxedge", 
                 base_learner = "standard_tree_regressor",
                 min_samples_split = 5, 
                 min_samples_leaf = 2,
                 max_depth = 2, 
                 order = 0, 
                 log_Xrange = True, 
                 random_state = 666,
                 parallel_jobs = 0, 
                 V = 2,
                 r_range_low = 0,
                 r_range_up = 1,
                 lamda = 0.01, 
                 search_number = 10,
                 threshold = 0
                 ):
        
        self.n_estimators = n_estimators
        self.max_features = max_features
        self.max_samples = max_samples
        self.rho = rho
        self.ensemble_parallel = ensemble_parallel
        self.splitter = splitter
        self.base_learner = base_learner
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_depth = max_depth
        self.order=order
        self.log_Xrange = log_Xrange
        self.random_state = random_state
        self.parallel_jobs = parallel_jobs
        self.V = V
        self.r_range_up =r_range_up
        self.r_range_low =r_range_low
        self.lamda=lamda
        self.search_number = search_number
        self.threshold = threshold


        self.trees = []

        
    def fit(self, X, y):
        
        assert self.ensemble_parallel * self.parallel_jobs == 0
        
        f_hat = np.zeros(X.shape[0])
        for i in range(self.n_estimators):
            bootstrap_idx = np.random.choice(X.shape[0], int(X.shape[0] * self.max_samples))
            
            self.trees.append(BASE_LEARNER[self.base_learner](splitter = self.splitter, 
                                             min_samples_split = self.min_samples_split,
                                             min_samples_leaf = self.min_samples_leaf,
                                             max_depth = self.max_depth,
                                             order = self.order, 
                                             log_Xrange = self.log_Xrange, 
                                             random_state = i,
                                             parallel_jobs = self.parallel_jobs,
                                             V = self.V,
                                             r_range_low = self.r_range_low,
                                             r_range_up = self.r_range_up,
                                             lamda = self.lamda,
                                             max_features = self.max_features,
                                             search_number = self.search_number,
                                             threshold = self.threshold))
            
            self.trees[i].fit(X[bootstrap_idx], (y-f_hat)[bootstrap_idx])
            f_hat += self.rho * self.trees[i].predict(X)
        
    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in [ "n_estimators" ,'min_samples_split', "max_features", "max_samples", "rho"
                    "splitter", "min_samples_leaf", "max_depth", "order", "V", 
                    "r_range_low", "r_range_up", "lamda",
                    "search_number", "threshold"]:
            value = getattr(self, key, None)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out
    
    
    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)


        for key, value in params.items():
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))
            setattr(self, key, value)
            valid_params[key] = value

        return self
    
    def predict(self, X):
        if self.ensemble_parallel == 0:
            y_hat = np.zeros(X.shape[0])
            for i in range(self.n_estimators):
                y_hat += self.rho * self.trees[i].predict(X)
            return y_hat

        else:
            with Pool(min(self.ensemble_parallel, self.n_estimators)) as pp:
                y_hat = pp.map(pred_parallel, [(self.trees[i],X) for i in range(self.n_estimators)])
            y_hat = self.rho* np.array(y_hat).sum(axis = 0)
            return y_hat
    
    

    
    
class GradientBoostingTreeRegressor(BaseGradientBoosting):
    """Standard gradient boosted regression trees using naive estimator.
    """
    def __init__(self, n_estimators = 20, 
                 max_features = 1.0, 
                 max_samples = 1.0,
                 rho = 0.1,
                 ensemble_parallel = 0,
                 splitter = "maxedge", 
                 min_samples_split = 5, 
                 min_samples_leaf = 2,
                 max_depth = 2, 
                 order = 0, 
                 log_Xrange = True, 
                 random_state = 666,
                 parallel_jobs = 0, 
                 V = 2,
                 r_range_low = 0,
                 r_range_up = 1,
                 lamda = 0.01, 
                 search_number = 10,
                 threshold = 0):
        super(GradientBoostingTreeRegressor, self).__init__(n_estimators = n_estimators,
                                                 max_features = max_features,
                                                 max_samples = max_samples,
                                                 rho = rho,
                                                 ensemble_parallel = ensemble_parallel,
                                                 splitter = splitter,
                                                 base_learner = "standard_tree_regressor", 
                                                 min_samples_split = min_samples_split,
                                                 min_samples_leaf = min_samples_leaf,
                                                 max_depth = max_depth, 
                                                 order = order,
                                                 log_Xrange = log_Xrange, 
                                                 random_state = random_state,
                                                 parallel_jobs = parallel_jobs,
                                                 V = V,
                                                 r_range_low = r_range_low,
                                                 r_range_up = r_range_up,
                                                 lamda = lamda,
                                                 search_number = search_number,
                                                 threshold = threshold)
        
    def score(self, X, y):
        """Reture the regression score, i.e. MSE.
        """
        return -MSE(self.predict(X),y)
    
    
    
    
class GradientBoostingExtraTreeRegressor(BaseGradientBoosting):
    """Gradient boosted extrapolated regression trees using extrapolated trees.
    """
    def __init__(self, n_estimators = 20, 
                 max_features = 1.0, 
                 max_samples = 1.0,
                 rho = 0.1,
                 ensemble_parallel = 0,
                 splitter = "maxedge", 
                 min_samples_split = 5, 
                 min_samples_leaf = 2,
                 max_depth = 2, 
                 order = 0, 
                 log_Xrange = True, 
                 random_state = 666,
                 parallel_jobs = 0, 
                 V = 2,
                 r_range_low = 0,
                 r_range_up = 1,
                 lamda = 0.01, 
                 search_number = 10,
                 threshold = 0):
        super(GradientBoostingExtraTreeRegressor, self).__init__(n_estimators = n_estimators,
                                                 max_features = max_features,
                                                 max_samples = max_samples,
                                                 rho = rho,
                                                 ensemble_parallel = ensemble_parallel,
                                                 splitter = splitter,
                                                 base_learner = "extra_tree_regressor", 
                                                 min_samples_split = min_samples_split,
                                                 min_samples_leaf = min_samples_leaf,
                                                 max_depth = max_depth, 
                                                 order = order,
                                                 log_Xrange = log_Xrange, 
                                                 random_state = random_state,
                                                 parallel_jobs = parallel_jobs,
                                                 V = V,
                                                 r_range_low = r_range_low,
                                                 r_range_up = r_range_up,
                                                 lamda = lamda,
                                                 search_number = search_number,
                                                 threshold = threshold)
        
    def score(self, X, y):
        """Reture the regression score, i.e. MSE.
        """
        return -MSE(self.predict(X),y)













from .tree import StandardTreeRegressor, ExtraTreeRegressor#, ExtraTreeClassifier, ExtraTreeDensityEstimator

#__all__ = ["ExtraTreeRegressor", "ExtraTreeClassifier", "ExtraTreeDensityEstimator"]

__all__ = ["StandardTreeRegressor",
           "ExtraTreeRegressor"]


import numpy as np
from multiprocessing import Pool


def assign_parallel_jobs(input_tuple):
    """Parallel instrumental function

    Parameters
    ----------
    input_tuple : (int, estimator object, array of (n_samples_, dim))
        The tuple for parallelization.
    
    Returns
    -------
    idx : int
        Index for parallel jobs.
        
    prediction : array-like of shape (n_sample_, ) 
        Prediction.
    """
    idx, node_object, X =input_tuple
    return idx, node_object.predict(X)


_TREE_LEAF = -1
_TREE_UNDEFINED = -2

class TreeStruct(object):
    """ Basic Binary Tree Structure.
    
    
    Parameters
    ----------
    
    log_Xrange : bool
        If True, the boundary of each cell is recorded. 
        
    n_samples : int
        Number of samples.
    
    n_features : int
        Number of dimensions.
        
    Attributes
    ----------
    node_count : int
        Log of number of nodes.
        
    left_child : list
        Log of all left children nodes.
        
    right_child : list
        Log of all right children nodes.
        
    feature : array of int
        Log of splitting dimensions.
    
    threshold : array of float 
        Log of splitting points.
        
    n_node_samples : array of int
        Log of number of points in each node.
        
    node_range : list
        Log of boundary of nodes. 
    
    """
    def __init__(self, n_samples, n_features, log_Xrange=True):
        
        # Base tree statistic
        self.n_samples = n_samples
        self.n_features = n_features
        self.node_count = 0
        # Base inner tree struct
        self.left_child = []
        self.right_child = []
        self.feature = []
        self.threshold = []
        self.n_node_samples = []
        self.leaf_ids = []
        self.leafnode_fun = {}  
        
        # If log_Xrange is True, the range of each node is also logged.  
        self.log_Xrange = log_Xrange
       
        if log_Xrange == True:
            self.node_range = []
            
    def _node_append(self):
        """Add None to each logs as placeholders. 
        """
        self.left_child.append(None)
        self.right_child.append(None)
        self.feature.append(None)
        self.threshold.append(None)
        self.n_node_samples.append(None) 
        if self.log_Xrange == True:
            self.node_range.append(None)  
            
    def _add_node(self, parent, is_left, is_leaf, feature, threshold, n_node_samples, node_range=None):
        """Add a new node. 
        """
        self._node_append()
        node_id = self.node_count
        self.n_node_samples[node_id] = n_node_samples
        if self.log_Xrange == True:
            self.node_range[node_id] = node_range.copy()
            
        # record children status in parent nodes
        if parent != _TREE_UNDEFINED:
            if is_left:
                self.left_child[parent] = node_id
            else:
                self.right_child[parent] = node_id
        # record current node status
        if is_leaf:
            self.left_child[node_id] = _TREE_LEAF
            self.right_child[node_id] = _TREE_LEAF
            self.feature[node_id] = _TREE_UNDEFINED
            self.threshold[node_id] = _TREE_UNDEFINED
            self.leaf_ids.append(node_id)  
        else:
            # left_child and right_child will be set later
            self.feature[node_id] = feature
            self.threshold[node_id] = threshold
        self.node_count += 1
        return node_id
    
    def _node_info_to_ndarray(self):
        """Turn each logs into arrays. 
        """
        self.left_child = np.array(self.left_child, dtype=np.int32)
        self.right_child = np.array(self.right_child, dtype=np.int32)
        self.feature = np.array(self.feature, dtype=np.int32)
        self.threshold = np.array(self.threshold, dtype=np.float64)
        self.n_node_samples = np.array(self.n_node_samples, dtype=np.int32)
        self.leaf_ids = np.array(self.leaf_ids, dtype=np.int32)
        if self.log_Xrange == True:
            self.node_range = np.array(self.node_range, dtype=np.float64)
            
    def apply(self, X):
        """Get node ids.
        """
        return self._apply_dense(X)
    
    def _apply_dense(self, X):
        """Get node ids.
        """
        n = X.shape[0]
        result_nodeid = np.zeros(n, dtype=np.int32)
        for i in range(n):
            node_id = 0
            while self.left_child[node_id] != _TREE_LEAF:
                if X[i, self.feature[node_id]] < self.threshold[node_id]:
                    node_id = self.left_child[node_id]
                else:
                    node_id = self.right_child[node_id]
            result_nodeid[i] = node_id
        return  result_nodeid  
    
    def predict(self, X):
        """Predict for each x in X. 
        """
        node_affi = self.apply(X)
        y_predict_hat = np.zeros(X.shape[0])
        for leaf_id in self.leaf_ids:
            idx = node_affi == leaf_id
            
            y_predict_hat[idx] = self.leafnode_fun[leaf_id].predict(X[idx])
            
        return y_predict_hat
    
    def get_info(self, x):
        """Get the extrapolation information. 
        """
        assert len(x.shape) == 2
        node_affi = self.apply(x)[0]
        
        return self.leafnode_fun[node_affi].get_info(x)
       
    
    def predict_parallel(self, X, parallel_jobs):
        """Predict for each x in X using parallel computing
        """
        node_affi = self.apply(X)
        y_predict_hat = np.zeros(X.shape[0])
        
        if parallel_jobs == "auto":
            njobs = len(self.leaf_ids)
            #print("using {} threads".format(njobs))
        else:
            njobs = parallel_jobs
            
        with Pool(njobs) as p:
            result= p.map(assign_parallel_jobs, [ (leaf_id, self.leafnode_fun[leaf_id], X[node_affi == leaf_id]) for leaf_id in self.leaf_ids] )

        for return_vec in result:
            idx = node_affi == return_vec[0]
            y_predict_hat[idx]= return_vec[1]
            
        return y_predict_hat
    
    

class RecursiveTreeBuilder(object):
    """ Recursively build a binary tree structure.
    
    
    Parameters
    ----------
    splitter : splitter object
        
    Estimator : estimator object
        
    min_samples_split : int
        The minimum number of samples required to split an internal node.
    
    min_samples_leaf : int
        The minimum number of samples required in the subnodes to split an internal node.
    
    max_depth : int
        Maximum depth of the individual regression estimators.
        
    order : int > 0
        Extrapolation order.
    
        
    V : int or "auto"
        Parameter for homothetic estimation. If int, the estimations are taken at 
        i/V, i = 1, ..., V. If auto, it is set to max(n_samples * 2^(- 2 - max_depth), 5). 
        
    r_range_low : float in [0, 1]
        Lower bound of homothetic ratio to consider.
    
    r_range_up : float in [0, 1], > r_range_low
        Upper bound of homothetic ratio to consider.        
    
    lamda : float in [0,infty)
        Ridge regularization parameter. 
        
    
    """
    def __init__(self, splitter, 
                 Estimator, 
                 min_samples_split, 
                 min_samples_leaf,
                 max_depth,
                 order,
                 V,
                r_range_up,
                r_range_low,
                lamda):
        
        # about splitter
        self.splitter = splitter
        # about estimator
        self.Estimator = Estimator
        # about recursive splits
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_depth = max_depth
        self.order=order
        
        self.r_range_up=r_range_up
        self.r_range_low = r_range_low
        
        self.V = V
        self.lamda = lamda

        
    def build(self, tree, X, Y, X_range = None):
        """Grow the tree.
        """
        num_samples = X.shape[0]
        stack = []
        # prepare for stack [X, node_range, parent_status, left_node_status, depth]
        stack.append([X, Y,X_range, _TREE_UNDEFINED, _TREE_UNDEFINED, 0])
        while len(stack) != 0:
            dt_X, dt_Y, node_range, parent, is_left, depth = stack.pop()
            n_node_samples = dt_X.shape[0]
            # judge whether dt should be splitted or not
            if n_node_samples <= self.min_samples_split:
                is_leaf = True
            else:
      
                n_node_unique_samples = np.unique(np.hstack([dt_X,dt_Y.reshape(-1,1)]), axis=0).shape[0]
                
                # if too deep or contains too less samples, no split
                if depth >= self.max_depth or n_node_unique_samples <= self.min_samples_split:
                    is_leaf = True
                else:
                    rd_dim, rd_split = self.splitter(dt_X, node_range , dt_Y )
                    
                    
                    # if not proper split point (when criterion reduction diminishes)
                    if rd_split is None:
                        is_leaf = True
                    # pruning when the sub nodes contains few samples
                    elif (dt_X[:,rd_dim] >= rd_split).sum() < self.min_samples_leaf or (dt_X[:, rd_dim] < rd_split).sum() < self.min_samples_leaf:
                        is_leaf = True
                    
                    # if not any above, split
                    else:
                        is_leaf = False
                    
            # we will apply splits in non-leaf nodes
            if not is_leaf:
                node_id = tree._add_node(parent, is_left, is_leaf, rd_dim, rd_split, n_node_samples, node_range)
            else:
                node_id = tree._add_node(parent, is_left, is_leaf, None, None, n_node_samples, node_range)
                
                #print([(x<=node_range[1]).all() for x in dt_X])
                tree.leafnode_fun[node_id] = self.Estimator(node_range, 
                                                            num_samples,
                                                            dt_X, 
                                                            dt_Y,
                                                            self.order,
                                                            self.V,
                                                            self.r_range_up, 
                                                            self.r_range_low,
                                                            self.lamda
                                                           )
        
                tree.leafnode_fun[node_id].fit()
                
            # begin branching if the node is not leaf
            if not is_leaf:
                # update node range status
                if node_range is not None:
                    node_range_right = node_range.copy()
                    node_range_left = node_range.copy()
                    node_range_right[0, rd_dim] = rd_split
                    node_range_left[1, rd_dim] = rd_split
                else:
                    node_range_right = node_range_left = None
                    
                # push right child on stack
                right_idx = dt_X[:,rd_dim] >= rd_split
                dt_X_right = dt_X[right_idx]
                dt_Y_right = dt_Y[right_idx]
                stack.append([dt_X_right, dt_Y_right , node_range_right, node_id, False, depth+1])
                
                # Push left child on stack
                left_idx = ~right_idx
                dt_X_left = dt_X[left_idx]
                dt_Y_left= dt_Y[left_idx]
                stack.append([dt_X_left, dt_Y_left, node_range_left, node_id, True, depth+1])
                
               
                
        tree._node_info_to_ndarray()
        
        



import numpy as np
from numba import njit



@njit
def extrapolate_regression(dt_X, dt_Y, X_extra, X_range, order, 
                              r_range_low, r_range_up, V, lamda):
    """Compute the extrapolated result based on training samples in the cell.

    Parameters
    ----------
    dt_X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    X_extra : array-like of shape (dim_, )
        Extrapolated points.
    
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
    
    order : int > 0
        Extrapolation order.
    
    r_range_low : float in [0, 1]
        Lower bound of homothetic ratio to consider.
    
    r_range_up : float in [0, 1], > r_range_low
        Upper bound of homothetic ratio to consider.
        
    V : int
        Parameter for homothetic estimation. The estimations are taken at 
        i/V, i = 1, ..., V.
    
    lamda : float, >0
        Ridge regularization parameter. 
    
    Returns
    -------
    weights : array-like of shape (order, 1)
        Learnt coefficients of polynomials, weights[0,0] is the final extrapolated 
        estimation. 
        
    all_ratio : array-like of shape (n_sample_, )
        The sorted vector of homothetic ratios associated to all points in 
        the cell. 
        
    all_y_hat : array-like of shape (n_sample_, )
        The sorted vector of homothetic estimations associated to all points in 
        the cell. 
    
    all_ratio : array-like of shape (V, )
        The sorted vector of selected homothetic ratios. 
        
    all_y_hat : array-like of shape (V, )
        The sorted vector of selected homothetic estimations. 

    """
    n_pts = dt_X.shape[0]
    ratio_vec = np.zeros(n_pts)
    
    
    # standardize the cell to compute homothetic ratios
    for idx_X, X in enumerate(dt_X):
        centralized = X - X_extra
        for d in range(X_extra.shape[0]):
            positive_len = X_range[1,d] - X_extra[d]
            negative_len = X_extra[d] - X_range[0,d]
            
            if centralized[d] > 0:
                centralized[d] /= positive_len
            elif centralized[d] < 0:
                centralized[d] /= negative_len
         
        
        ratio_X = np.abs(centralized).max() 
        ratio_vec[idx_X] = ratio_X

    idx_sorted_by_ratio = np.argsort(ratio_vec)  
    
    #### all sorted ratio
    sorted_ratio = ratio_vec[idx_sorted_by_ratio]
    all_ratio = sorted_ratio.ravel()
    
    #### all sorted y
    sorted_y = dt_Y[idx_sorted_by_ratio]
    sorted_y_hat = np.zeros((n_pts,1))
    for k in range(n_pts):
        sorted_y_hat[k,0] = np.mean(sorted_y[ :(k+1)])
    all_y_hat = sorted_y_hat.ravel()
    
    # compute the homothetic estimations for selected sequence
    index_by_r = np.zeros(V)
    for t in range(V,0,-1):
        if_less_than_r = np.where(sorted_ratio <= t/V)[0]
        if len(if_less_than_r) <= 4:
            index_by_r[t-1] = 0
        else:
            index_by_r[t-1] = if_less_than_r.max()
    
    index_by_r = index_by_r[index_by_r > 0]    
    index_by_r = np.array([int(i) for i in index_by_r])

    sorted_y_hat = sorted_y_hat[index_by_r]
    sorted_ratio = sorted_ratio[index_by_r]
    
    # rule out points outside of [r_range_low, r_range_up]
    ratio_range_idx_up = sorted_ratio <= r_range_up
    ratio_range_idx_low  = sorted_ratio >= r_range_low
    ratio_range_idx = ratio_range_idx_up * ratio_range_idx_low
    sorted_ratio = sorted_ratio[ratio_range_idx]
    sorted_y_hat = sorted_y_hat[ratio_range_idx]
  
    # solve the least square problem
    ratio_mat = np.zeros((sorted_ratio.shape[0], order+1))
    i=0
    while(i < sorted_ratio.shape[0]):
        r = sorted_ratio[i] 
        for j in range(order +1):
            ratio_mat[i,j] = r**j 
        i+=1
        
    
    id_matrix = np.eye( ratio_mat.shape[1] )

    ratio_mat_T = np.ascontiguousarray(ratio_mat.T)
    ratio_mat = np.ascontiguousarray(ratio_mat)
    RTR = np.ascontiguousarray(ratio_mat_T @ ratio_mat+ id_matrix * lamda)
    RTR_inv = np.ascontiguousarray(np.linalg.inv(RTR))
    sorted_y_hat = np.ascontiguousarray(sorted_y_hat)
    

    return (RTR_inv @ ratio_mat_T @ sorted_y_hat ), all_ratio, all_y_hat, sorted_ratio, sorted_y_hat.ravel() 
   
    
   

class NaiveRegressionEstimator(object):
    """ Naive Regression Estimator
    
    The standard estimator which averages labels in the cell.
    
    
    Parameters
    ----------
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
        
    num_samples : int
        Number of samples in the cell.
        
    dt_X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    V : int
        Parameter for homothetic estimation. The estimations are taken at 
        i/V, i = 1, ..., V.
    
    order : int > 0
        Extrapolation order.
    
    r_range_low : float in [0, 1]
        Lower bound of homothetic ratio to consider.
    
    r_range_up : float in [0, 1], > r_range_low
        Upper bound of homothetic ratio to consider.
        
    
    lamda : float, >0
        Ridge regularization parameter. 
        
    Attributes
    ----------
    y_hat : float
        The final estimation, i.e. mean of dt_Y.
    """
    def __init__(self, 
                 X_range, 
                 num_samples, 
                 dt_X, 
                 dt_Y, 
                 order = None,
                 V = None,
                 r_range_up = None,
                 r_range_low = None,
                lamda = None):
        
        self.dt_Y = dt_Y
        self.n_node_samples = dt_X.shape[0]
        self.X_range = X_range
        
        
    def fit(self):
        if self.n_node_samples != 0:
            self.y_hat = self.dt_Y.mean()
        else:
            self.y_hat = 0
        
    def predict(self, test_X):
        y_predict = np.full(test_X.shape[0], self.y_hat)
        return y_predict
    

class ExtraRegressionEstimator(object):
    """ Extrapolation Regression Estimator
  
    
    Parameters
    ----------
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
        
    num_samples : int
        Number of samples in the cell.
        
    dt_X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    V : int
        Parameter for homothetic estimation. The estimations are taken at 
        i/V, i = 1, ..., V.
    
    order : int > 0
        Extrapolation order.
    
    r_range_low : float in [0, 1]
        Lower bound of homothetic ratio to consider.
    
    r_range_up : float in [0, 1], > r_range_low
        Upper bound of homothetic ratio to consider.
        
    
    lamda : float, >0
        Ridge regularization parameter. 
        
   
    """
    def __init__(self, 
                 X_range, 
                 num_samples, 
                 dt_X,
                 dt_Y,
                 order,
                 V = 1,
                 r_range_up=1,
                 r_range_low=0,
                lamda=0.01,
                ):
        
        self.X_range = X_range
        self.dim = X_range.shape[1]
        self.dt_X = dt_X
        self.dt_Y = dt_Y
        self.order = order
        self.lamda = lamda
        self.n_node_samples = dt_X.shape[0]
        self.r_range_up = r_range_up
        self.r_range_low = r_range_low
        self.V = V
    
    def fit(self):
        self.y_hat = None
        
    def predict(self, test_X):
        
        """Prediction function

        Parameters
        ----------
        test_X : array-like of shape (n_test_, dim_)
            An array of test points in the cell.
        
        Returns
        -------
        y_predict : array-like of shape (n_test_, )
            The prediction
        
        """
        
        assert self.V!=0
        
        if len(test_X)==0:
            return np.array([])

        pre_vec=[]

        # for each test x, compute extrapolated estimation
        for X in test_X:
            pred_weights, _, _, _, _ = extrapolate_regression(self.dt_X,
                                                                 self.dt_Y, 
                                                                 X, 
                                                                 self.X_range, 
                                                                 self.order,
                                                                 self.r_range_low, 
                                                                 self.r_range_up,
                                                                 self.V, 
                                                                 self.lamda)
            pre_vec.append(pred_weights[0,0])
            
        y_predict=np.array(pre_vec)
        return y_predict
    
    def get_info(self, x):
        """Query the extrapolation information

        Parameters
        ----------
        x : array-like of shape (dim_, 1)
            The extrapolation point.
        
        Returns
        -------
        weights : array-like of shape (order, 1)
            Learnt coefficients of polynomials, weights[0,0] is the final extrapolated 
            estimation. 
            
        all_ratio : array-like of shape (n_sample_, )
            The sorted vector of homothetic ratios associated to all points in 
            the cell. 
            
        all_y_hat : array-like of shape (n_sample_, )
            The sorted vector of homothetic estimations associated to all points in 
            the cell. 
        
        all_ratio : array-like of shape (V, )
            The sorted vector of selected homothetic ratios. 
            
        all_y_hat : array-like of shape (V, )
            The sorted vector of selected homothetic estimations. 
        
        """
        
        assert self.V != 0
        assert len(x.shape) == 2
        x = x.ravel()
        

        pred_weights, all_r , all_y_hat , used_r, used_y_hat = extrapolate_regression(self.dt_X,
                                                                                         self.dt_Y,  
                                                                                         x, 
                                                                                         self.X_range, 
                                                                                         self.order,
                                                                                         self.r_range_low,
                                                                                         self.r_range_up,
                                                                                         self.V,
                                                                                         self.lamda)
        return pred_weights, all_r, all_y_hat, used_r, used_y_hat


import numpy as np 
from ExtraTree.ExtraTree import StandardTreeRegressor, ExtraTreeRegressor



def test_standard_tree_regressor():
    
    for splitter in ["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"]:
        for threshold in [0,0.01]:
            for parallel_jobs in [0,5]:
                
                np.random.seed(666)
                X_train = np.random.rand(200).reshape(-1,2)
                X_test = np.random.rand(200).reshape(-1,2)
                Y_train = np.ones(100)

                model = StandardTreeRegressor( splitter = splitter,
                                        min_samples_split = 5, 
                                        min_samples_leaf = 2,
                                        max_depth = 2, 
                                        log_Xrange = True, 
                                        random_state = 666,
                                        parallel_jobs = parallel_jobs, 
                                        max_features = 1.0,
                                        search_number = 10,
                                        threshold = threshold)
                model.fit(X_train, Y_train)
                assert ((model.predict(X_test)-1)**2).mean()<0.1
    
    
def test_extra_tree_regressor():
    
    for splitter in ["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"]:
        for order in [0,1,5]:
            for threshold in [0,0.01]:
                for parallel_jobs in [0,5]:
                    for lamda in [0.001]:
                    
                        np.random.seed(666)
                        X_train = np.random.rand(200).reshape(-1,2)
                        X_test = np.random.rand(200).reshape(-1,2)
                        Y_train = np.ones(100)

                        model = ExtraTreeRegressor( splitter = splitter,
                                                min_samples_split = 5, 
                                                min_samples_leaf = 2,
                                                max_depth = 2, 
                                                order = order, 
                                                log_Xrange = True, 
                                                random_state = 666,
                                                parallel_jobs = parallel_jobs, 
                                                V = 10,
                                                r_range_low = 0,
                                                r_range_up = 1,
                                                lamda = lamda, 
                                                max_features = 1.0,
                                                search_number = 10,
                                                threshold = threshold)
                        model.fit(X_train, Y_train)

                      
                        assert ((model.predict(X_test)-1)**2).mean()<0.1
                    
                    



import numpy as np


def insample_ssq(y):
    """In sample sum of squares.
    """
    return np.var(y)*len(y)



def mse(X, dt_Y, d, split):
    """Compute MSE decrease for one pair of dimension and split point.

    Parameters
    ----------
    
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
 
    d : int in 0, ..., dim - 1
        The splitting dimension.
    
    split : float
        The splitting point.
    
    Returns
    -------
    
    mse_reduction : float
        The decrease of mse. 

    """
    before_mse = insample_ssq( dt_Y )
    after_mse = insample_ssq( dt_Y[ X[:,d] < split ] ) + insample_ssq( dt_Y[ X[:,d] >= split ] )
    
    return (before_mse - after_mse) / len(dt_Y)


def gini(X, dt_Y, d, split):
    """Compute GINI decrease for one pair of dimension and split point.

    Parameters
    ----------
    
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
 
    d : int in 0, ..., dim - 1
        The splitting dimension.
    
    split : float
        The splitting point.
    
    Returns
    -------
    
    gini_reduction : float
        The decrease of GINI. 

    """
    return None


from .ExtraTree import StandardTreeRegressor, ExtraTreeRegressor#, ExtraTreeClassifier, ExtraTreeDensityEstimator
from .ExtraEnsemble import GradientBoostingTreeRegressor, GradientBoostingExtraTreeRegressor, StandardForestRegressor, ExtraForestRegressor

__all__ = ["StandardTreeRegressor",
           "ExtraTreeRegressor",
           "GradientBoostingTreeRegressor", 
           "GradientBoostingExtraTreeRegressor", 
           "StandardForestRegressor", 
           "ExtraForestRegressor",
           ]


import numpy as np 
from ExtraTree.ExtraTree import StandardTreeRegressor, ExtraTreeRegressor
from sklearn.model_selection import GridSearchCV



def test_standard_tree_regressor():
    
                
    np.random.seed(666)
    X_train = np.random.rand(200).reshape(-1,2)
    X_test = np.random.rand(200).reshape(-1,2)
    Y_train = np.ones(100)

    parameters = {"max_depth":[2, 4, 6, 8],
                  "splitter":["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"],
                  "threshold":[0, 0.01],
                  }
    
    cv_model_standard_tree = GridSearchCV(estimator = StandardTreeRegressor(), param_grid = parameters, cv = 3) 
    cv_model_standard_tree.fit(X_train, Y_train)
    
    model_standard_tree = cv_model_standard_tree.best_estimator_
    assert (model_standard_tree.predict(X_test)==1).all()
    
    
def test_extra_tree_regressor():
                    
    np.random.seed(666)
    X_train = np.random.rand(200).reshape(-1,2)
    X_test = np.random.rand(200).reshape(-1,2)
    Y_train = np.ones(100)

    parameters = {"max_depth":[2, 4, 6, 8],
                  "splitter":["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"],
                  "threshold":[0, 0.01],
                  "order":[0, 1, 5],
                  "lamda":[0.001]
                  }
    
    cv_model_extra_tree = GridSearchCV(estimator = ExtraTreeRegressor(), param_grid = parameters, cv = 3) 
    cv_model_extra_tree.fit(X_train, Y_train)
    
    model_extra_tree = cv_model_extra_tree.best_estimator_
    assert ((model_extra_tree.predict(X_test)-1)**2).mean()<0.01
    
    
    



def test_standard_tree_regressor_noise():
    
                
    np.random.seed(666)
    X_train = np.random.rand(200).reshape(-1,2)
    X_test = np.random.rand(200).reshape(-1,2)
    Y_train = np.ones(100) + np.random.normal(scale = 0.1, size = 100)

    parameters = {"max_depth":[2, 4, 6, 8],
                  "splitter":["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"],
                  "threshold":[0, 0.01],
                  }
    
    cv_model_standard_tree = GridSearchCV(estimator = StandardTreeRegressor(), param_grid = parameters, cv = 3) 
    cv_model_standard_tree.fit(X_train, Y_train)
    
    model_standard_tree = cv_model_standard_tree.best_estimator_
    assert (((model_standard_tree.predict(X_test)-1)**2).mean() < 0.15).all()
    
    
def test_extra_tree_regressor_noise():
                    
    np.random.seed(666)
    X_train = np.random.rand(200).reshape(-1,2)
    X_test = np.random.rand(200).reshape(-1,2)
    Y_train = np.ones(100) + np.random.normal(scale = 0.1, size = 100)

    parameters = {"max_depth":[2, 4, 6, 8],
                  "splitter":["purely", "midpoint", "maxedge", "msereduction", "msemaxedge"],
                  "threshold":[0, 0.01],
                  "order":[0, 1, 5],
                  "lamda":[0.001]
                  }
    
    cv_model_extra_tree = GridSearchCV(estimator = ExtraTreeRegressor(), param_grid = parameters, cv = 3) 
    cv_model_extra_tree.fit(X_train, Y_train)
    
    model_extra_tree = cv_model_extra_tree.best_estimator_
    assert (((model_extra_tree.predict(X_test)-1)**2).mean() < 0.15).all()

                      
                        
                    
                    


                      
                        
                    
                    



from .boosting import GradientBoostingTreeRegressor, GradientBoostingExtraTreeRegressor
from .forest import StandardForestRegressor, ExtraForestRegressor

__all__ = ["GradientBoostingTreeRegressor", "GradientBoostingExtraTreeRegressor",
           "StandardForestRegressor", "ExtraForestRegressor"]


import numpy as np
from ._criterion import gini, mse

criterion_func = {"gini":gini,
                  "mse":mse}

class PurelyRandomSplitter(object):    
    """Purely random splitter class.

    Parameters
    ----------
    random_state : int
        Random state for dimension subsampling and splitting.
    
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
    
    
    Returns
    -------
    rd_dim : int in 0, ..., dim - 1
        The splitting dimension.
        
    rd_split : float
        The splitting point.

    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = None, threshold = None):
        self.random_state = random_state
        np.random.seed(self.random_state)
        
    def __call__(self, X, X_range, dt_Y = None):
        n_node_samples, dim = X.shape
        
        # randomly choose dimension and split point
        rd_dim = np.random.randint(0, dim)
        rddim_min = X_range[0, rd_dim]
        rddim_max = X_range[1, rd_dim]
        rd_split = np.random.uniform(rddim_min, rddim_max)
        return rd_dim, rd_split
    
    
class MidPointRandomSplitter(object):
    """Random mid-point splitter class.

    Parameters
    ----------
    random_state : int
        Random state for dimension subsampling and splitting.
    
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
    
    
    Returns
    -------
    rd_dim : int in 0, ..., dim - 1
        The splitting dimension.
        
    rd_split : float
        The splitting point.

    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = None, threshold = None):
        self.random_state = random_state
        np.random.seed(self.random_state)
        self.max_features = max_features
        
    def __call__(self, X, X_range, dt_Y = None):
        n_node_samples, dim = X.shape
        
        # randomly choose a dimension and split at mid-point
        rd_dim = np.random.randint(0, dim)
        rddim_min = X_range[0, rd_dim]
        rddim_max = X_range[1, rd_dim]
        rd_split = (rddim_min+ rddim_max)/2
        return rd_dim, rd_split
    
    
class MaxEdgeRandomSplitter(object):
    """Random max-edge splitter class.

    Parameters
    ----------
    random_state : int
        Random state for dimension subsampling and splitting.
    
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
    
    
    Returns
    -------
    rd_dim : int in 0, ..., dim - 1
        The splitting dimension.
        
    rd_split : float
        The splitting point.

    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = None, threshold = None):
        self.random_state = random_state
        self.max_features = max_features
        np.random.seed(self.random_state)
        
    def __call__(self, X, X_range ,dt_Y = None):
        n_node_samples, dim = X.shape
        
        # randomly choose among the dimensions with longest edge
        edge_ratio = X_range[1] - X_range[0]
        
        # sub-sample a subset of dimensions
        subsampled_idx = np.random.choice(edge_ratio.shape[0], int(np.ceil(edge_ratio.shape[0] * self.max_features)),replace = False)
        rd_dim = np.random.choice(np.where(edge_ratio[subsampled_idx] == edge_ratio[subsampled_idx].max())[0])
        rddim_min = X_range[0, rd_dim]
        rddim_max = X_range[1, rd_dim]
        rd_split = (rddim_min + rddim_max)/2
        return rd_dim, rd_split
    
    
    
    
    
    
    
    
    
    
class GainReductionSplitter(object):
    """Abstract information gain based splitter class.

    Parameters
    ----------
    random_state : int
        Random state for dimension subsampling and splitting.
    
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
    
    
    Returns
    -------
    rd_dim : int in 0, ..., dim - 1
        The splitting dimension.
        
    rd_split : float
        The splitting point.

    """
    def __init__(self, criterion, random_state = None, max_features = 1.0, search_number = 10, threshold = None):
        self.random_state = random_state
        np.random.seed(self.random_state)
        self.max_features = max_features
        self.search_number = search_number
        self.compute_criterion_reduction = criterion_func[criterion]
        self.threshold = threshold
        
    def __call__(self, X, X_range, dt_Y):
        n_node_samples, dim = X.shape
        
        # sub-sample a subset of dimensions
        subsampled_idx = np.random.choice(dim, int(np.ceil( dim * self.max_features) ), replace = False)

        max_criterion_reduction = np.inf
        split_dim = None
        split_point = None
        
        # search for dimension and split point with maximum criterion reduction
        for d in subsampled_idx:
            
            dt_X_dim_unique = np.unique(X[:,d])
            sorted_split_point = np.unique( np.quantile( dt_X_dim_unique, [(2 * i + 1)/(2 * self.search_number) for i in range(self.search_number) ] ) )
            
            for split in sorted_split_point:
                
                # compute criterion
                criterion_reduction = self.compute_criterion_reduction(X, dt_Y, d, split)
            
                # hault if reduction is small
                if criterion_reduction < max_criterion_reduction and criterion_reduction >= self.threshold:
                    
                    max_criterion_reduction = criterion_reduction
                    split_dim = d
                    split_point = split
            
                

        return split_dim, split_point
    

class MSEReductionSplitter(GainReductionSplitter):
    """MSE reduction splitter class.
    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = 10, threshold = None):
        super(MSEReductionSplitter, self).__init__( criterion = "mse", 
                                                   random_state = random_state, 
                                                   max_features = max_features, 
                                                   search_number = search_number,
                                                   threshold = threshold)
    
    
class GINIReductionSplitter(GainReductionSplitter):
    """GINI reduction splitter class.
    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = 10, threshold = None):
        super(GINIReductionSplitter, self).__init__( criterion = "gini", 
                                                   random_state = random_state, 
                                                   max_features = max_features, 
                                                   search_number = search_number,
                                                   threshold = threshold)
        
        
        










class GainReductionMaxEdgeSplitter(object):
    """Abstract information gain based mid-point splitter class.

    Parameters
    ----------
    random_state : int
        Random state for dimension subsampling and splitting.
    
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    X : array-like of shape (n_sample_, dim_)
        An array of points in the cell.
    
    dt_Y : array-like of shape (n_sample_, )
        An array of labels in the cell.
        
    X_range : array-like of shape (2, dim_)
        Boundary of the cell, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
    
    
    Returns
    -------
    rd_dim : int in 0, ..., dim - 1
        The splitting dimension.
        
    rd_split : float
        The splitting point.

    """
    def __init__(self, criterion, random_state = None, max_features = 1.0, search_number = None, threshold = None):
        self.random_state = random_state
        np.random.seed(self.random_state)
        self.max_features = max_features
        self.compute_criterion_reduction = criterion_func[criterion]
        self.threshold = threshold
        
    def __call__(self, X, X_range, dt_Y):
        n_node_samples, dim = X.shape
        
        
        edge_ratio = X_range[1] - X_range[0]
        
        # sub-sample a subset of dimensions
        subsampled_idx = np.random.choice(edge_ratio.shape[0], int(np.ceil(edge_ratio.shape[0] * self.max_features)),replace = False)
        max_edges = np.where(edge_ratio[subsampled_idx] == edge_ratio[subsampled_idx].max())[0]
        

        max_criterion_reduction = np.inf
        split_dim = None
        split_point = None
        
        # search for dimension with maximum criterion reduction
        for rd_dim in max_edges:
            
            
            split = ( X_range[1,rd_dim] + X_range[0,rd_dim])/2
                
            # compute criterion
            criterion_reduction = self.compute_criterion_reduction(X, dt_Y, rd_dim, split)
            
            # hault if reduction is small
            if criterion_reduction < max_criterion_reduction and criterion_reduction >= self.threshold:
                
                max_criterion_reduction = criterion_reduction
                split_dim = rd_dim
                split_point = split
            
                

        return split_dim, split_point
    


class MSEReductionMaxEdgeSplitter(GainReductionSplitter):
    """MSE reduction mid-point splitter class.
    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = None, threshold = None):
        super(MSEReductionMaxEdgeSplitter, self).__init__( criterion = "mse", 
                                                   random_state = random_state, 
                                                   max_features = max_features, 
                                                   search_number = search_number,
                                                   threshold = threshold)
    
    
class GINIReductionMaxEdgeSplitter(GainReductionSplitter):
    """GINI reduction mid-point splitter class.
    """
    def __init__(self, random_state = None, max_features = 1.0, search_number = None, threshold = None):
        super(GINIReductionMaxEdgeSplitter, self).__init__( criterion = "gini", 
                                                   random_state = random_state, 
                                                   max_features = max_features, 
                                                   search_number = search_number,
                                                   threshold = threshold)
        
        
        
  


from setuptools import find_packages, setup



import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")




with open('requirements.txt') as inn:
    requirements = inn.read().splitlines()

setup(
    name = 'ExtraTree',

    version = '0.0.3',

    description = 'Extrapolated Tree',

    long_description = long_description,

    long_description_content_type = "text/markdown",

    url = 'https://github.com/Karlmyh/ExtraTree',

    author = "Yuheng Ma",

    author_email = "yma@ruc.edu.cn",
    
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    
    
    packages = find_packages(),

    python_requires = '>=3',
    
    install_requires = requirements,
)



import numpy as np
from sklearn.metrics import mean_squared_error as MSE

from ._tree import TreeStruct, RecursiveTreeBuilder
from ._splitter import PurelyRandomSplitter, MidPointRandomSplitter, MaxEdgeRandomSplitter, MSEReductionSplitter, GINIReductionSplitter, MSEReductionMaxEdgeSplitter, GINIReductionMaxEdgeSplitter

from ._estimator import NaiveRegressionEstimator, ExtraRegressionEstimator



SPLITTERS = {"purely": PurelyRandomSplitter,
             "midpoint": MidPointRandomSplitter, 
             "maxedge": MaxEdgeRandomSplitter, 
             "msereduction": MSEReductionSplitter,
             "ginireduction": GINIReductionSplitter,
             "msemaxedge": MSEReductionMaxEdgeSplitter,
             "ginimaxedge": GINIReductionMaxEdgeSplitter
             }

ESTIMATORS = {"naive_regression": NaiveRegressionEstimator,
              "extra_regression": ExtraRegressionEstimator,
              }


class BaseRecursiveTree(object):
    """ Abstact Recursive Tree Structure.
    
    
    Parameters
    ----------
    splitter : splitter keyword in SPLITTERS
        Splitting scheme
        
    estimator : estimator keyword in ESTIMATORS
        Estimation scheme
        
    min_samples_split : int
        The minimum number of samples required to split an internal node.
    
    min_samples_leaf : int
        The minimum number of samples required in the subnodes to split an internal node.
    
    max_depth : int
        Maximum depth of the individual regression estimators.
        
    order : int > 0
        Extrapolation order.
    
    log_Xrange : bool
        If True, the points in each cell is recorded. 
        
    random_state : int
        Random state for building the tree.
        
    parallel_jobs : int
        If 0, no parallel. If positive, the parallization is conducted in parallel_jobs threads.
        
    V : int or "auto"
        Parameter for homothetic estimation. If int, the estimations are taken at 
        i/V, i = 1, ..., V. If auto, it is set to max(n_samples * 2^(- 2 - max_depth), 5). 
        
    r_range_low : float in [0, 1]
        Lower bound of homothetic ratio to consider.
    
    r_range_up : float in [0, 1], > r_range_low
        Upper bound of homothetic ratio to consider.        
    
    lamda : float in [0,infty)
        Ridge regularization parameter. 
        
    max_features : float in (0,1]
        Proportion of dimensions to consider when splitting. 
        
    search_number : int
        Number of points to search on when looking for best split point.
        
    threshold : float in [0, infty]
        Threshold for haulting when criterion reduction is too small.
        
    Attributes
    ----------
    n_samples : int
        Number of samples.
    
    dim : int
        Dimension of covariant.
        
    X_range : array-like of shape (2, dim_)
        Boundary of the support, X_range[0, d] and X_range[1, d] stands for the
        lower and upper bound of d-th dimension.
        
    tree_ : binary tree object defined in _tree.py
    
    """
    def __init__(self, 
                 splitter = None, 
                 estimator = None, 
                 min_samples_split = None,
                 min_samples_leaf = None,
                 max_depth = None, 
                 order = None, 
                 log_Xrange = None, 
                 random_state = None,
                 parallel_jobs = None,
                 V = None,
                 r_range_up = None,
                 r_range_low = None,
                 lamda = None,
                 max_features = None,
                 search_number = None,
                 threshold = None
                ):
        self.splitter = splitter
        self.estimator = estimator
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_depth = max_depth
        self.order = order
        self.log_Xrange = log_Xrange
        self.random_state = random_state
        self.parallel_jobs = parallel_jobs
        self.V = V
        self.r_range_up = r_range_up
        self.r_range_low = r_range_low
        self.lamda = lamda
        self.max_features = max_features
        self.search_number = search_number
        self.threshold = threshold
             
    def fit(self, X, Y, X_range = "unit"):
        
        self.n_samples, self.dim = X.shape
        
        # check the boundary
        if X_range == "unit":
            X_range = np.array([np.zeros(self.dim),np.ones(self.dim)])
        if X_range is None:
            X_range = np.zeros(shape = (2, X.shape[1]))
            X_range[0] = X.min(axis = 0) - 0.01 * (X.max(axis = 0) - X.min(axis = 0))
            X_range[1] = X.max(axis = 0) + 0.01 * (X.max(axis = 0) - X.min(axis = 0))
        self.X_range = X_range


        # set V
        if self.V == "auto":
            V =  max(5, int(X.shape[0]* 2**(-self.max_depth-2)))
        else:
            V = self.V
        


        # begin
        SPLITTERS[self.splitter]
        splitter = SPLITTERS[self.splitter](self.random_state, self.max_features, self.search_number, self.threshold)
        
        Estimator = ESTIMATORS[self.estimator]
        # initiate a tree structure
        self.tree_ = TreeStruct(self.n_samples, self.dim, self.log_Xrange)
        # recursively build the tree
        builder = RecursiveTreeBuilder(splitter, 
                                       Estimator, 
                                       self.min_samples_split, 
                                       self.min_samples_leaf,
                                       self.max_depth, 
                                       self.order,
                                       V,
                                      self.r_range_up,
                                      self.r_range_low,
                                      self.lamda)
        builder.build(self.tree_, X, Y, X_range)
        
        return self
        
    def apply(self, X):
        """Reture the belonging cell ids. 
        """
        return self.tree_.apply(X)
    
    def get_info(self, x):
        """Query the extrapolation information
        """
        return self.tree_.get_info(x)
    
    def get_node_idx(self,X):
        """Reture the belonging cell ids. 
        """
        return self.apply(X)
    
    def get_node(self,X):
        """Reture the belonging node. 
        """
        return [self.tree_.leafnode_fun[i] for i in self.get_node_idx(X)]
    
    def get_all_node(self):
        """Reture all nodes. 
        """
        return list(self.tree_.leafnode_fun.values())
    
    def predict(self, X):
        if self.parallel_jobs != 0:
            #print("we are using parallel computing!")
            y_hat = self.tree_.predict_parallel(X, self.parallel_jobs)
        else:
            y_hat = self.tree_.predict(X)
        
        # check boundary
        check_lowerbound = (X - self.X_range[0] >= 0).all(axis = 1)
        check_upperbound = (X - self.X_range[1] <= 0).all(axis = 1)
        is_inboundary = check_lowerbound * check_upperbound
        # assign 0 to points outside the boundary
        y_hat[np.logical_not(is_inboundary)] = 0
        return y_hat
    
    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in ['min_samples_split',"min_samples_leaf", "max_depth", "order", "V", 
                    "splitter", "r_range_low", "r_range_up", "lamda", "max_features",
                    "search_number", "threshold" ]:
            value = getattr(self, key, None)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out
    
    
    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)


        for key, value in params.items():
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))
            setattr(self, key, value)
            valid_params[key] = value

        return self
    
    


class StandardTreeRegressor(BaseRecursiveTree):
    """Standard regression tree using naive estimator.
    """
    def __init__(self, splitter = "maxedge", 
                 min_samples_split = 5, 
                 min_samples_leaf = 2,
                 max_depth = 2, 
                 order = 0, 
                 log_Xrange = True, 
                 random_state = 666,
                 parallel_jobs = 0, 
                 V = 2,
                 r_range_low = 0,
                 r_range_up = 1,
                 lamda = 0.01, 
                 max_features = 1.0,
                 search_number = 10,
                 threshold = 0):
        super(StandardTreeRegressor, self).__init__(splitter = splitter,
                                             estimator = "naive_regression", 
                                             min_samples_split = min_samples_split,
                                             min_samples_leaf = min_samples_leaf,
                                             max_depth = max_depth, 
                                             log_Xrange = log_Xrange, 
                                             random_state = random_state,
                                             parallel_jobs = parallel_jobs,
                                             max_features = max_features,
                                             search_number = search_number,
                                             threshold = threshold)
        
    def score(self, X, y):
        """Reture the regression score, i.e. MSE.
        """
        return -MSE(self.predict(X),y)
    
    

class ExtraTreeRegressor(BaseRecursiveTree):
    """Extrapolated regression tree using extrapolated estimator.
    """
    def __init__(self, splitter = "maxedge", 
                 min_samples_split = 5, 
                 min_samples_leaf = 2,
                 max_depth = 2, 
                 order = 0, 
                 log_Xrange = True, 
                 random_state = 666,
                 parallel_jobs = 0, 
                 V = 2,
                 r_range_low = 0,
                 r_range_up = 1,
                 lamda = 0.01, 
                 max_features = 1.0,
                 search_number = 10,
                 threshold = 0):
        super(ExtraTreeRegressor, self).__init__(splitter = splitter,
                                             estimator = "extra_regression", 
                                             min_samples_split = min_samples_split,
                                             min_samples_leaf = min_samples_leaf,
                                             max_depth = max_depth, 
                                             order = order,
                                             log_Xrange = log_Xrange, 
                                             random_state = random_state,
                                             parallel_jobs = parallel_jobs,
                                             V = V,
                                             r_range_low = r_range_low,
                                             r_range_up = r_range_up,
                                             lamda = lamda,
                                             max_features = max_features,
                                             search_number = search_number,
                                             threshold = threshold)
        
    def score(self, X, y):
        """Reture the regression score, i.e. MSE.
        """
        return -MSE(self.predict(X),y)

